import itertools
import warnings
from datetime import datetime
from typing import Iterable, Optional

import numpy as np
import scipy.sparse as spr
from sklearn.base import clone
from sklearn.decomposition._nmf import (NMF, MiniBatchNMF, _initialize_nmf,
                                        _update_coordinate_descent)
from sklearn.exceptions import NotFittedError
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.utils import check_array
from sklearn.utils.validation import check_non_negative

from turftopic.base import Encoder

NOT_MATCHING_ERROR = (
    "Document embedding dimensionality ({n_dims}) doesn't match term embedding dimensionality ({n_word_dims}). "
    + "Perhaps you are using precomputed embeddings but forgot to pass an encoder to your model. "
    + "Try to initialize the model with the encoder you used for computing the embeddings."
)


def batched(iterable, n: int) -> Iterable[list[str]]:
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := list(itertools.islice(it, n)):
        yield batch


def fit_timeslice(
    X,
    W,
    H,
    tol=1e-4,
    max_iter=200,
    l1_reg_W=0,
    l1_reg_H=0,
    l2_reg_W=0,
    l2_reg_H=0,
    verbose=0,
    shuffle=False,
    random_state=None,
):
    """Fits topic_term_matrix based on a precomputed document_topic_matrix.
    This is used to get temporal components in dynamic KeyNMF.
    """
    Ht = check_array(H.T, order="C")
    if random_state is None:
        rng = np.random.mtrand._rand
    else:
        rng = np.random.RandomState(random_state)
    for n_iter in range(1, max_iter + 1):
        violation = 0.0
        violation += _update_coordinate_descent(
            X.T, Ht, W, l1_reg_H, l2_reg_H, shuffle, rng
        )
        if n_iter == 1:
            violation_init = violation
        if violation_init == 0:
            break
        if verbose:
            print("violation:", violation / violation_init)
        if violation / violation_init <= tol:
            if verbose:
                print("Converged at iteration", n_iter + 1)
            break
    return W, Ht.T, n_iter


class SBertKeywordExtractor:
    def __init__(
        self, top_n: int, encoder: Encoder, vectorizer: CountVectorizer
    ):
        self.top_n = top_n
        self.encoder = encoder
        self.vectorizer = vectorizer
        self.key_to_index: dict[str, int] = {}
        self.term_embeddings: Optional[np.ndarray] = None

    @property
    def is_encoder_promptable(self) -> bool:
        prompts = getattr(self.encoder, "prompts", None)
        if prompts is None:
            return False
        if ("query" in prompts) and ("passage" in prompts):
            return True

    @property
    def n_vocab(self) -> int:
        return len(self.key_to_index)

    def _add_terms(self, new_terms: list[str]):
        for term in new_terms:
            self.key_to_index[term] = self.n_vocab
        if not self.is_encoder_promptable:
            term_encodings = self.encoder.encode(new_terms)
        else:
            term_encodings = self.encoder.encode(
                new_terms, prompt_name="passage"
            )
        if self.term_embeddings is not None:
            self.term_embeddings = np.concatenate(
                (self.term_embeddings, term_encodings), axis=0
            )
        else:
            self.term_embeddings = term_encodings

    def batch_extract_keywords(
        self,
        documents: list[str],
        embeddings: Optional[np.ndarray] = None,
    ) -> list[dict[str, float]]:
        if not len(documents):
            return []
        if embeddings is None:
            if not self.is_encoder_promptable:
                embeddings = self.encoder.encode(documents)
            else:
                embeddings = self.encoder.encode(
                    documents, prompt_name="query"
                )
        if len(embeddings) != len(documents):
            raise ValueError(
                "Number of documents doesn't match number of embeddings."
            )
        keywords = []
        vectorizer = clone(self.vectorizer)
        document_term_matrix = vectorizer.fit_transform(documents)
        batch_vocab = vectorizer.get_feature_names_out()
        new_terms = list(set(batch_vocab) - set(self.key_to_index.keys()))
        if len(new_terms):
            self._add_terms(new_terms)
        total = embeddings.shape[0]
        for i in range(total):
            terms = document_term_matrix[i, :].todense()
            embedding = embeddings[i].reshape(1, -1)
            mask = terms > 0
            if not np.any(mask):
                keywords.append(dict())
                continue
            important_terms = np.ravel(np.asarray(mask))
            word_embeddings = [
                self.term_embeddings[self.key_to_index[term]]
                for term in batch_vocab[important_terms]
            ]
            if self.term_embeddings.shape[1] != embeddings.shape[1]:
                raise ValueError(
                    NOT_MATCHING_ERROR.format(
                        n_dims=embeddings.shape[1],
                        n_word_dims=self.term_embeddings.shape[1],
                    )
                )
            sim = cosine_similarity(embedding, word_embeddings).astype(
                np.float64
            )
            sim = np.ravel(sim)
            kth = min(self.top_n, len(sim) - 1)
            top = np.argpartition(-sim, kth)[:kth]
            top_words = batch_vocab[important_terms][top]
            top_sims = [sim for sim in sim[top] if sim > 0]
            keywords.append(dict(zip(top_words, top_sims)))
        return keywords


class KeywordNMF:
    def __init__(
        self,
        n_components: int,
        seed: Optional[int] = None,
        top_n: Optional[int] = None,
    ):
        self.n_components = n_components
        self.key_to_index: dict[str, int] = {}
        self.index_to_key: list[str] = []
        self.top_n = top_n
        # n_components * n_vocab
        self.components: Optional[np.ndarray] = None
        self.seed = seed
        self.temporal_components: Optional[np.ndarray] = None
        self.temporal_importance_: Optional[np.ndarray] = None

    def prune_keywords(self, keywords: dict[str, float]) -> dict[str, float]:
        """If there are more keywords than allowed, this prunes them."""
        if (self.top_n is None) or (self.top_n >= len(keywords)):
            return keywords
        words, similarities = zip(*keywords.items())
        similarities = np.array(similarities)
        selected = np.argsort(-similarities)[: self.top_n]
        items = [(words[i], similarities[i]) for i in selected]
        return dict(items)

    @property
    def n_vocab(self) -> int:
        return len(self.index_to_key)

    def _add_word_components(self, X: spr.csr_matrix):
        """Initializes components for novel vocabulary."""
        _, H = _initialize_nmf(X, self.n_components, random_state=self.seed)
        if self.components is None:
            self.components = H
        else:
            n_new = X.shape[1] - self.components.shape[1]
            if n_new:
                self.components = np.concatenate(
                    (self.components, H[:, -n_new:]), axis=1
                )
        if self.temporal_components is not None:
            n_new = X.shape[1] - self.temporal_components.shape[-1]
            if n_new:
                new_comps = H[:, -n_new:]
                new_comps = np.broadcast_to(
                    new_comps,
                    (self.temporal_components.shape[0], *new_comps.shape),
                )
                self.temporal_components = np.concatenate(
                    (self.temporal_components, new_comps), axis=-1
                )

    def vectorize(
        self, keywords: list[dict[str, float]], fitting: bool = False
    ) -> spr.csr_array:
        indices = []
        indptr = [0]
        values = []
        for k in keywords:
            k = self.prune_keywords(k)
            for w, v in k.items():
                # Adding vocab item if missing
                if (w not in self.key_to_index) and fitting:
                    self.key_to_index[w] = self.n_vocab
                    self.index_to_key.append(w)
                if w in self.key_to_index:
                    indices.append(self.key_to_index[w])
                    values.append(v)
            indptr.append(len(indices))
        shape = (len(indptr) - 1, self.n_vocab)
        document_term_matrix = spr.csr_matrix(
            (values, indices, indptr), shape=shape
        )
        return document_term_matrix

    def fit_transform(self, keywords: list[dict[str, float]]) -> np.ndarray:
        X = self.vectorize(keywords, fitting=True)
        check_non_negative(X, "NMF (input X)")
        W, H = _initialize_nmf(X, self.n_components, random_state=self.seed)
        W, H, self.n_iter = NMF(
            self.n_components, init="custom", random_state=self.seed
        )._fit_transform(X, W=W, H=H, update_H=True)
        self.components = H.astype(X.dtype)
        return W

    def transform(self, keywords: list[dict[str, float]]):
        if self.components is None:
            raise NotFittedError(
                "Can't transform() if the model has not been fitted."
            )
        X = self.vectorize(keywords, fitting=False)
        check_non_negative(X, "NMF (input X)")
        W, _, _ = NMF(
            self.n_components, init="custom", random_state=self.seed
        )._fit_transform(X, W=None, H=self.components, update_H=False)
        return W.astype(X.dtype)

    def partial_fit(self, keyword_batch: list[dict[str, float]]):
        X = self.vectorize(keyword_batch, fitting=True)
        try:
            check_non_negative(X, "NMF (input X)")
            self._add_word_components(X)
            W, _ = _initialize_nmf(
                X, self.n_components, random_state=self.seed
            )
            _minibatchnmf = MiniBatchNMF(
                self.n_components, init="custom", random_state=self.seed
            ).partial_fit(X, W=W, H=self.components)
            self.components = _minibatchnmf.components_.astype(X.dtype)
        except ValueError as e:
            warnings.warn(f"Batch failed with error: {e}, skipping.")
            return self
        return self

    def fit_transform_dynamic(
        self,
        keywords: list[dict[str, float]],
        time_labels: np.ndarray,
        time_bin_edges: list[datetime],
    ) -> np.ndarray:
        self.time_bin_edges = time_bin_edges
        n_bins = len(time_bin_edges) - 1
        document_term_matrix = self.vectorize(keywords, fitting=True)
        check_non_negative(document_term_matrix, "NMF (input X)")
        document_topic_matrix, H = _initialize_nmf(
            document_term_matrix,
            self.n_components,
            random_state=self.seed,
        )
        document_topic_matrix, H, self.n_iter = NMF(
            self.n_components, init="custom", random_state=self.seed
        )._fit_transform(
            document_term_matrix, W=document_topic_matrix, H=H, update_H=True
        )
        self.components = H.astype(document_term_matrix.dtype)
        n_comp, n_vocab = self.components.shape
        self.temporal_components = np.zeros(
            (n_bins, n_comp, n_vocab), dtype=document_term_matrix.dtype
        )
        self.temporal_importance_ = np.zeros((n_bins, n_comp))
        for label in np.unique(time_labels):
            idx = np.nonzero(time_labels == label)
            X = document_term_matrix[idx]
            W = document_topic_matrix[idx]
            _, H = _initialize_nmf(
                X, self.components.shape[0], random_state=self.seed
            )
            _, H, _ = fit_timeslice(X, W, H, random_state=self.seed)
            self.temporal_components[label] = H
            topic_importances = np.squeeze(np.asarray(W.sum(axis=0)))
            self.temporal_importance_[label] = topic_importances
        return document_topic_matrix

    def partial_fit_dynamic(
        self,
        keyword_batch: list[dict[str, float]],
        time_labels: np.ndarray,
        time_bin_edges: list[datetime],
    ) -> np.ndarray:
        if self.temporal_components is None:
            self.fit_transform_dynamic(
                keyword_batch, time_labels, time_bin_edges
            )
        else:
            document_term_matrix = self.vectorize(keyword_batch, fitting=True)
            check_non_negative(document_term_matrix, "NMF (input X)")
            self._add_word_components(document_term_matrix)
            document_topic_matrix = self.transform(keyword_batch)
            _minibatchnmf = MiniBatchNMF(
                self.n_components, init="custom", random_state=self.seed
            ).partial_fit(
                document_term_matrix,
                W=document_topic_matrix,
                H=self.components,
            )
            self.components = _minibatchnmf.components_.astype(
                document_term_matrix.dtype
            )
            document_topic_matrix = self.transform(keyword_batch)
            for label in np.unique(time_labels):
                idx = np.nonzero(time_labels == label)
                X = document_term_matrix[idx]
                W = document_topic_matrix[idx]
                _minibatchnmf = MiniBatchNMF(
                    self.n_components, init="custom", random_state=self.seed
                ).partial_fit(
                    X,
                    W=W,
                    H=self.temporal_components[label],
                )
                self.temporal_components[label] = _minibatchnmf.components_
                topic_importances = np.squeeze(np.asarray(W.sum(axis=0)))
                self.temporal_importance_[label] += topic_importances



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

import numpy as np
from rich.console import Console
from rich.tree import Tree

from turftopic.base import ContextualModel

COLOR_PER_LEVEL = [
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_green",
    "bright_red",
    "bright_yellow",
    "cyan",
    "magenta",
    "blue",
    "white",
]


def _tree_plot(hierarchy: TopicNode):
    """Plots hierarchy with Plotly as a Tree"""
    try:
        import igraph as ig
        import plotly.express as px
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "You will need to install plotly and igraph to use hierarchical plotting functionality."
        ) from e

    def traverse(h, nodes, edges, parent=None):
        nodes.append(h)
        if parent is not None:
            edges.append([parent._simple_desc, h._simple_desc])
        if h.children is not None:
            for child in h.children:
                traverse(child, nodes, edges, parent=h)

    def word_table(h):
        entries = []
        words = h.get_words(top_k=10)
        for word, imp in words:
            entries.append(f"<b>{word}</b>: <i>{imp:.2f}</i>")
        return " <br> ".join(entries)

    nodes = []
    edges = []
    for child in hierarchy.children:
        traverse(child, nodes, edges)
    node_names = [node._simple_desc for node in nodes]
    node_to_idx = {node_name: idx for idx, node_name in enumerate(node_names)}
    edges_idx = [
        [node_to_idx[start], node_to_idx[end]] for start, end in edges
    ]
    tables = [word_table(node) for node in nodes]
    graph = ig.Graph(len(nodes), edges=edges_idx, directed=True)
    layout = graph.layout("rt")
    layout.rotate(-90)
    x, y = np.array(layout.coords).T
    xmin, xmax = np.min(x), np.max(x)
    xpad = (xmax - xmin) * 0.35
    fig = px.scatter(x=x, y=y, text=node_names, template="plotly_white")
    fig = fig.update_traces(
        customdata=[[table] for table in tables],
        hovertemplate="<b>%{text}</b> <br> <br> %{customdata[0]}",
    )
    fig = fig.update_traces(marker=dict(size=20, color="rgba(0,0,0,0.2)"))
    fig = fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
    fig = fig.update_yaxes(showgrid=False, visible=False, zeroline=False)
    fig = fig.update_xaxes(
        showgrid=False,
        visible=False,
        zeroline=False,
        range=(xmin - xpad, xmax + xpad),
    )
    for start, end in edges_idx:
        fig.add_shape(
            type="line",
            xref="x",
            yref="y",
            x0=x[start],
            y0=y[start],
            x1=x[end],
            y1=y[end],
            opacity=0.2,
        )
    return fig


@dataclass
class TopicNode:
    """Node for a topic in a topic hierarchy.

    Parameters
    ----------
    model: ContextualModel
        Underlying topic model, which the hierarchy is based on.
    path: tuple[int], default ()
        Path that leads to this node from the root of the tree.
    word_importance: ndarray of shape (n_vocab), default None
        Importance of each word in the vocabulary for given topic.
    document_topic_vector: ndarray of shape (n_documents), default None
        Importance of the topic in all documents in the corpus.
    children: list[TopicNode], default None
        List of subtopics within this topic.
    """

    model: ContextualModel
    path: tuple[int] = ()
    word_importance: Optional[np.ndarray] = None
    document_topic_vector: Optional[np.ndarray] = None
    children: Optional[list[TopicNode]] = None

    @classmethod
    def create_root(
        cls,
        model: ContextualModel,
        components: np.ndarray,
        document_topic_matrix: np.ndarray,
    ) -> TopicNode:
        """Creates root node from a topic models' components and topic importances in documents."""
        children = []
        n_components = components.shape[0]
        for i, comp, doc_top in zip(
            range(n_components), components, document_topic_matrix.T
        ):
            children.append(
                cls(
                    model,
                    path=(i,),
                    word_importance=comp,
                    document_topic_vector=doc_top,
                    children=None,
                )
            )
        return TopicNode(
            model,
            path=(),
            word_importance=None,
            document_topic_vector=None,
            children=children,
        )

    @property
    def level(self) -> int:
        """Indicates how deep down the hierarchy the topic is."""
        return len(self.path)

    def get_words(self, top_k: int = 10) -> list[tuple[str, float]]:
        """Returns top words and words importances for the topic.

        Parameters
        ----------
        top_k: int, default 10
            Number of top words to return.

        Returns
        -------
        list[tuple[str, float]]
            List of word, importance pairs.
        """
        if (self.word_importance is None) or (
            self.document_topic_vector
        ) is None:
            return []
        idx = np.argpartition(-self.word_importance, top_k)[:top_k]
        order = np.argsort(-self.word_importance[idx])
        idx = idx[order]
        imp = self.word_importance[idx]
        words = self.model.get_vocab()[idx]
        return list(zip(words, imp))

    @property
    def description(self) -> str:
        """Returns a high level description of the topic with its path in the tree
        and top words."""
        if not len(self.path):
            path = "Root"
        else:
            path = ".".join([str(idx) for idx in self.path])
        words = []
        for word, imp in self.get_words(top_k=10):
            words.append(word)
        concat_words = ", ".join(words)
        color = COLOR_PER_LEVEL[min(self.level, len(COLOR_PER_LEVEL) - 1)]
        stylized = f"[{color} bold]{path}[/]: [italic]{concat_words}[/]"
        console = Console()
        with console.capture() as capture:
            console.print(stylized, end="")
        return capture.get()

    @property
    def _simple_desc(self) -> str:
        if not len(self.path):
            path = "Root"
        else:
            path = ".".join([str(idx) for idx in self.path])
        words = []
        for word, imp in self.get_words(top_k=5):
            words.append(word)
        concat_words = ", ".join(words)
        return f"{path}: {concat_words}"

    def _build_tree(self, tree: Tree = None, top_k: int = 10) -> Tree:
        if tree is None:
            tree = Tree(self.description)
        else:
            tree = tree.add(self.description)
        if self.children is not None:
            for child in self.children:
                child._build_tree(tree)
        return tree

    def __str__(self):
        tree = self._build_tree(top_k=10)
        console = Console()
        with console.capture() as capture:
            console.print(tree)
        return capture.get()

    def __repr__(self):
        return str(self)

    def clear(self):
        """Deletes children of the given node."""
        self.children = None
        return self

    def __getitem__(self, index: int):
        if self.children is None:
            raise IndexError("Current node is a leaf and has not children.")
        return self.children[index]

    def divide(self, n_subtopics: int, **kwargs):
        """Divides current node into smaller subtopics.
        Only works when the underlying model is a divisive hierarchical model.

        Parameters
        ----------
        n_subtopics: int
            Number of topics to divide the topic into.
        """
        try:
            self.children = self.model.divide_topic(
                node=self, n_subtopics=n_subtopics, **kwargs
            )
        except AttributeError as e:
            raise AttributeError(
                "Looks like your model is not a divisive hierarchical model."
            ) from e
        return self

    def divide_children(self, n_subtopics: int, **kwargs):
        """Divides all children of the current node to smaller topics.
        Only works when the underlying model is a divisive hierarchical model.

        Parameters
        ----------
        n_subtopics: int
            Number of topics to divide the topics into.
        """
        if self.children is None:
            raise ValueError(
                "Current Node is a leaf, children can't be subdivided."
            )
        for child in self.children:
            child.divide(n_subtopics, **kwargs)
        return self

    def plot_tree(self):
        """Plots hierarchy as an interactive tree in Plotly."""
        return _tree_plot(self)
